# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Create transformation instructions for transformation_performer.

Given quantization parameters, create a list of transformation instructions that
can then be used by transformation_performer. Includes necessary optimizations
"""

from collections.abc import Iterator
import dataclasses
from typing import Optional
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import constrained_ops_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


_OpQuantConstraint = common_utils.OpQuantConstraint
_QuantTransformation = qtyping.QuantTransformation


# When a tensor has no producer, we'll assign -1 to the producer field
# When a tensor is a graph output, we'll also include a -1 in the consumer list
def check_horizontal_optimization(
    param1: qtyping.OpToTensorParams,
    param2: qtyping.OpToTensorParams,
    index: int,
) -> bool:
  """check if horizontal optimization can be applied.

  check if two transformations at the same index (which belongs to two
  different
  OpToTensorParams) can be merged together.

  Args:
    param1: first parameters to be compared
    param2: second parameters to be compared
    index: the index for which the transformation will be compared

  Returns:
    True if the two transformations can be merged, False otherwise
  """
  if (
      isinstance(param1.parameters, qtyping.UniformQuantParams)
      and param1.parameters.hadamard is not None
  ):
    if (
        isinstance(param2.parameters, qtyping.UniformQuantParams)
        and param2.parameters.hadamard is not None
    ):
      return True
  return (
      param1.parameters == param2.parameters
      and len(param1.transformations) > index
      and len(param2.transformations) > index
      and param1.transformations[index] == param2.transformations[index]
  )


def check_dq_q_elimination(
    producer_inst: qtyping.TransformationInst,
    consumer_inst: qtyping.TransformationInst,
) -> bool:
  """Check if a pair of dequantize & quantize transformation can be eliminated.

  This can only happen when the dequantize & quantize have the same quant
  parameters and dequantize belongs to producer and quantize belongs to a
  consumer.

  Args:
    producer_inst: TransformationInst from producer.
    consumer_inst: TransformationInst from consumer.

  Returns:
    True if dequantize & quantize can be eliminated, False otherwise.
  """
  is_dequantize_in_producer = (
      producer_inst.transformation == qtyping.QuantTransformation.ADD_DEQUANTIZE
  )
  is_quantize_in_consumer = (
      consumer_inst.transformation == qtyping.QuantTransformation.ADD_QUANTIZE
  )
  is_same_parameters = producer_inst.parameters == consumer_inst.parameters
  return (
      is_dequantize_in_producer
      and is_quantize_in_consumer
      and is_same_parameters
  )


def check_replace_dq_q_with_rq(
    producer_inst: qtyping.TransformationInst,
    consumer_inst: qtyping.TransformationInst,
) -> bool:
  """Check if a pair of dequantize & quantize can be replaced by a requantize.

  This can only happen when the dequantize belongs to producer and quantize
  belongs to a consumer.

  Args:
    producer_inst: TransformationInst from producer.
    consumer_inst: TransformationInst from consumer.

  Returns:
    True if dequantize & quantize can be replaced, False otherwise.
    Note that we consider the case where DQ & Q can be eliminated as a false
    case.
  """
  is_dequantize_in_producer = (
      producer_inst.transformation == qtyping.QuantTransformation.ADD_DEQUANTIZE
  )
  is_quantize_in_consumer = (
      consumer_inst.transformation == qtyping.QuantTransformation.ADD_QUANTIZE
  )
  is_same_parameters = producer_inst.parameters == consumer_inst.parameters

  return (
      is_dequantize_in_producer
      and is_quantize_in_consumer
      and not is_same_parameters
  )


def check_dq_no_quant_elimination(
    producer_inst: qtyping.TransformationInst,
    consumer_inst: qtyping.TransformationInst,
) -> bool:
  """Check if a pair of dequantize & no quantize transformation can be eliminated.

  This can only happen when the dequantize belongs to producer and no quantize
  belongs to a consumer.

  Args:
    producer_inst: TransformationInst from producer.
    consumer_inst: TransformationInst from consumer.

  Returns:
    True if dequantize & no quantize can be eliminated, False otherwise.
  """
  is_dequantize_in_producer = (
      producer_inst.transformation == qtyping.QuantTransformation.ADD_DEQUANTIZE
  )
  is_no_quant_in_consumer = (
      consumer_inst.transformation == qtyping.QuantTransformation.NO_QUANTIZE
  )
  return is_dequantize_in_producer and is_no_quant_in_consumer


class TransformationInstructionsGenerator:
  """Generates transformation instructions from tensor quant params."""

  def __init__(self, float_tflite: Optional[str] = None):
    """Constructor.

    Args:
      float_tflite: the original TFlite model in bytearray or file path.
    """
    if float_tflite is None:
      self._tensor_name_to_graph_info: dict[
          str, TransformationInstructionsGenerator.TensorGraphInfo
      ] = {}
      self.flatbuffer_model: schema_py_generated.ModelT = ()
    else:
      self.flatbuffer_model = tfl_flatbuffer_utils.read_model(float_tflite)
      self._create_tensor_name_to_graph_info_map()
    self._same_as_input_scale_ops = (
        constrained_ops_utils.get_constrained_op_list(
            _OpQuantConstraint.SAME_AS_INPUT_SCALE
        )
    )
    self._same_as_output_scale_ops = (
        constrained_ops_utils.get_constrained_op_list(
            _OpQuantConstraint.SAME_AS_OUTPUT_SCALE
        )
    )

  @dataclasses.dataclass(frozen=True)
  class TensorGraphInfo:
    tensor_id: int
    subgraph_id: int
    producer: int
    consumers: list[int]

  def _tensor_info_generator(
      self, subgraph_id: int, subgraph: schema_py_generated.SubGraphT
  ) -> Iterator[tuple[str, TensorGraphInfo]]:
    """Generator function for tensor info.

    Args:
      subgraph_id: Index for the given subgraph,
      subgraph: Subgraph struct to generate tensor info on.

    Yields:
      A tuple of tensor_name and TensorGraphInfo.
    """
    for tensor_id, tensor in enumerate(subgraph.tensors):
      consumers = []
      for op_id, op in enumerate(subgraph.operators):
        # Some ops may use the same input tensor multiple times,
        # and we should handle each time independently.
        for op_input in op.inputs:
          if op_input == tensor_id:
            consumers.append(op_id)
      producer = -1
      for op_id, op in enumerate(subgraph.operators):
        if tensor_id in op.outputs:
          producer = op_id
          break
      if tensor_id in subgraph.outputs:
        consumers.insert(0, -1)
      tensor_info = self.TensorGraphInfo(
          tensor_id, subgraph_id, producer, consumers
      )
      tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
      yield tensor_name, tensor_info

  def _create_tensor_name_to_graph_info_map(self):
    """Create a mapping between tensor name and tensor info."""
    self._tensor_name_to_graph_info = {}
    # TODO: b/333607428 - support graph input & output
    for subgraph_id, subgraph in enumerate(self.flatbuffer_model.subgraphs):
      for tensor_name, tensor_info in self._tensor_info_generator(
          subgraph_id, subgraph
      ):
        self._tensor_name_to_graph_info[tensor_name] = tensor_info

  def _group_consumer_transformations(
      self, param: qtyping.TensorTransformationParams
  ) -> list[list[set[int]]]:
    """Group transformations between consumers into common groups.

    Args:
      param: TensorTransformationParams for a tensor

    Returns:
      A list of list of sets where the set represents indices of transformations
      that can be merged horizontally
      E.g:
        For the following consumer:
         [(1, [ADD_QUANTIZE, ADD_DEQUANTIZE], param1),
          (2, [ADD_QUANTIZE], param2),
          (3, [ADD_QUANTIZE], param1)]
        this function returns:
        [[{1, 2, 3}],
         [{1, 3}, {2}],
         [{1}]]

        Where the 0 depth list is the initial state, since all consumer comes
        from the same producer.
        In depth 1, the ADD_QUANTIZE in 1 & 3 can be merged, so they are in the
        same group
        In depth 2, there is only one transformation from 1, so there is only
        one group with 1 in there
    """
    if not param or not param.consumers:
      return []

    # consumer group contains indices of operations that can be horizontally
    # optimized together. The outermost list is the depth of the transformation
    # and the second list contains sets that represents the consumer indices
    # that can be grouped together at the given depth
    consumer_groups = [[set()]]
    # the max number of transformations applied before a particular consumer
    longest_trans_chain = 0
    for i, consumer_param in enumerate(param.consumers):
      consumer_groups[0][0].add(i)
      longest_trans_chain = max(
          longest_trans_chain, len(consumer_param.transformations)
      )

    #  looping over transformations of the same depth
    for transformation_depth in range(longest_trans_chain):
      next_depth_groups = []
      for consumer_param_index, consumer_param in enumerate(param.consumers):
        if len(consumer_param.transformations) > transformation_depth:
          for current_depth_groups in consumer_groups[transformation_depth]:
            if consumer_param_index in current_depth_groups:
              # if the transformation of the particular edge has been processed
              trans_assigned = False
              for new_group in next_depth_groups:
                # get an index in the existing group, any of them work since
                # they have the same quantization
                index = next(iter(new_group))
                if (
                    index in current_depth_groups
                    and check_horizontal_optimization(
                        param.consumers[index],
                        consumer_param,
                        transformation_depth,
                    )
                ):
                  new_group.add(consumer_param_index)
                  trans_assigned = True
                  break
              if not trans_assigned:
                next_depth_groups.append(set([consumer_param_index]))
      consumer_groups.append(next_depth_groups)
    return consumer_groups

  def _produce_transformation_for_vertical_opt(
      self,
      consumer_group: list[list[set[int]]],
      param: qtyping.TensorTransformationParams,
  ) -> list[qtyping.TransformationInst]:
    """Create a list of transformation rules available for vertical optimization.

    A consumer transformation is available to vertical transformation IFF it's
    the first transformation for a given consumer.

    This function relies on the consumer_group argument already being optimized
    for horizontal transformations.

    Args:
      consumer_group: a list of grouped indices for consumer transformationns
      param: a TensorTransformationParams for the tensor

    Returns:
      A list of transformation rules available for vertical optimization
    """
    tensor_info = self._tensor_name_to_graph_info[param.tensor_name]
    transformations_available_for_vertical_optimization = []
    # we start at 1 because consumer groups in index 0 is the inital state
    # and does not contain actual information
    if len(consumer_group) > 1:
      for group in consumer_group[1]:
        op_list = list(group)
        op_idx_list = []
        for index in op_list:
          op_idx_list.append(param.consumers[index].subgraph_op_id)
        transformations_available_for_vertical_optimization.append(
            qtyping.TransformationInst(
                param.consumers[op_list[0]].transformations[0],
                tensor_info.tensor_id,
                tensor_info.producer,
                op_idx_list,
                param.consumers[op_list[0]].parameters,
            )
        )
    return transformations_available_for_vertical_optimization

  def _produce_consumer_transformations_unavailable_for_vertical_opt(
      self,
      consumer_group: list[list[set[int]]],
      param: qtyping.TensorTransformationParams,
  ) -> list[qtyping.TransformationInst]:
    """Produce a list of consumer transformation that can't be used for vertical optimization.

    A consumer transformation is available to vertical optimization if and only
    if it's the first transformation for a given consumer.

    This function relies on the consumer_group argument already being optimized
    for horizontal transformations

    Args:
      consumer_group: a list of grouped indices for consumer transformationns
      param: a TensorTransformationParams for the tensor

    Returns:
      A list of transformation rules unavailable for vertical optimization
    """
    tensor_info = self._tensor_name_to_graph_info[param.tensor_name]
    other_consumer_transformations = []
    for transformation_idx in range(2, len(consumer_group)):
      for group in consumer_group[transformation_idx]:
        op_list = list(group)
        op_idx_list = []
        if (
            len(param.consumers[op_list[0]].transformations)
            <= transformation_idx - 1
        ):
          continue
        for index in op_list:
          op_idx_list.append(param.consumers[index].subgraph_op_id)
        other_consumer_transformations.append(
            qtyping.TransformationInst(
                param.consumers[op_list[0]].transformations[
                    transformation_idx - 1
                ],
                tensor_info.tensor_id,
                tensor_info.producer,
                op_idx_list,
                param.consumers[op_list[0]].parameters,
            )
        )
    return other_consumer_transformations

  def _apply_vertical_optimization(
      self,
      producer_trans_rule: qtyping.TransformationInst,
      consumer_trans_rules: list[qtyping.TransformationInst],
  ) -> list[qtyping.TransformationInst]:
    """Apply vertical optimization.

    There are two types of transformations we consider:
      1. when DQ & Q has the same parameter eliminate the operators and quantize
      the tensor only
      2. when DQ & Q has different parameters, then replace the DQ & Q with an
      RQ op

    vertical optimization can only happen with the last producer rules and the
    first consumer rules that are on the first.

    Args:
      producer_trans_rule: the last producer transformation rules.
      consumer_trans_rules: a list of consumer transformation rules that are
        avilable for vertical transformations.

    Returns:
      A list of transformations after vertical optimization has been applied,
      note producer transformation is included.
    """
    transformations = []
    for trans_rule in consumer_trans_rules:
      if check_dq_q_elimination(producer_trans_rule, trans_rule):
        for consumer_id in trans_rule.consumers:
          if consumer_id in producer_trans_rule.consumers:
            producer_trans_rule.consumers.remove(consumer_id)
        transformations.append(
            qtyping.TransformationInst(
                qtyping.QuantTransformation.QUANTIZE_TENSOR,
                trans_rule.tensor_id,
                trans_rule.producer,
                trans_rule.consumers,
                trans_rule.parameters,
            )
        )
        continue
      elif check_replace_dq_q_with_rq(producer_trans_rule, trans_rule):
        for consumer_id in trans_rule.consumers:
          producer_trans_rule.consumers.remove(consumer_id)
        transformations.append(
            qtyping.TransformationInst(
                qtyping.QuantTransformation.QUANTIZE_TENSOR,
                trans_rule.tensor_id,
                trans_rule.producer,
                trans_rule.consumers,
                producer_trans_rule.parameters,
            )
        )
        transformations.append(
            qtyping.TransformationInst(
                qtyping.QuantTransformation.ADD_QUANTIZE,
                trans_rule.tensor_id,
                trans_rule.producer,
                trans_rule.consumers,
                trans_rule.parameters,
            )
        )
        continue
      elif check_dq_no_quant_elimination(producer_trans_rule, trans_rule):
        for consumer_id in trans_rule.consumers:
          if consumer_id in producer_trans_rule.consumers:
            producer_trans_rule.consumers.remove(consumer_id)
        transformations.append(
            qtyping.TransformationInst(
                qtyping.QuantTransformation.ADD_DEQUANTIZE,
                trans_rule.tensor_id,
                trans_rule.producer,
                trans_rule.consumers,
                producer_trans_rule.parameters,
            )
        )
        continue
      else:
        transformations.append(trans_rule)
    if producer_trans_rule.consumers:
      transformations.insert(0, producer_trans_rule)
    return transformations

  def _remove_last_tensor_duplication(
      self, tensor_trans_insts: qtyping.TensorTransformationInsts
  ) -> None:
    """Remove the last tensor duplication so the original tensor can be reused."""
    instructions = tensor_trans_insts.instructions
    if not instructions:
      return
    for i in range(len(instructions) - 1, -1, -1):
      if (
          instructions[i].transformation
          == _QuantTransformation.DUPLICATE_TENSOR
      ):
        instructions.pop(i)
        return

  def _remove_unnecessary_buffer_duplication(
      self, tensor_trans_insts: qtyping.TensorTransformationInsts
  ) -> None:
    """Remove buffer duplications that comes after a tensor duplication.

    When a tensor is duplicated, a new buffer is created for it. Therefore,
    buffer duplication transformation that comes after it is unnecessary.

    Args:
      tensor_trans_insts: Transformation instructions for a tensor.
    """
    instructions = tensor_trans_insts.instructions
    if not instructions:
      return

    # Find all consumers that have a tensor duplication.
    consumers_with_tensor_duplication = set()
    for instr in instructions:
      if instr.transformation == _QuantTransformation.DUPLICATE_TENSOR:
        consumers_with_tensor_duplication.update(instr.consumers)
    if not consumers_with_tensor_duplication:
      return

    # Remove a buffer duplication that comes with a tensor duplication.
    for i in range(len(instructions) - 1, -1, -1):
      instr = instructions[i]
      if (
          instr.transformation == _QuantTransformation.DUPLICATE_BUFFER
          and consumers_with_tensor_duplication.issuperset(instr.consumers)
      ):
        instructions.pop(i)

  def _is_valid_quantize_requantize_pair(
      self,
      instr_0: qtyping.TransformationInst,
      instr_1: qtyping.TransformationInst,
  ) -> bool:
    """Checks if the two instructions form a valid quantize and requantize pair."""
    return (
        instr_0.transformation == _QuantTransformation.QUANTIZE_TENSOR
        and instr_1.transformation == _QuantTransformation.ADD_QUANTIZE
        and instr_0.consumers == instr_1.consumers
    )

  def _is_op_constrained(
      self, subgraph_id: int, op_index: int
  ) -> bool:
    """Checks if the op has same as input or output scale constraints."""
    op_name = tfl_flatbuffer_utils.get_op_name_by_index(
        self.flatbuffer_model, subgraph_id, op_index
    )
    return (
        op_name in self._same_as_input_scale_ops
        or op_name in self._same_as_output_scale_ops
    )

  def _are_quant_params_compatible(
      self,
      params_0: qtyping.UniformQuantParams,
      params_1: qtyping.UniformQuantParams,
  ) -> bool:
    """Checks if quant params are the same except for the scale and zero point."""
    ignore_set = {"scale", "zero_point"}
    for field_info in dataclasses.fields(qtyping.UniformQuantParams):
      field_name = field_info.name
      if field_name in ignore_set:
        continue
      if getattr(params_0, field_name) != getattr(params_1, field_name):
        return False
    return True

  def _eliminate_requantization_for_nonconstrained_provider(
      self, tensor_trans_insts: qtyping.TensorTransformationInsts
  ) -> None:
    """Removes requantization for tensors with a non-constrained provider.

    Fuses [QUANTIZE_TENSOR, ADD_QUANTIZE] instructions when a tensor has a
    provider op without same as input/ouput scale constrains. Quant params from
    the second instruction are copied to the first one and ADD_QUANTIZE is
    removed.

    Args:
      tensor_trans_insts: Transformation instructions for a tensor.
    """
    instructions = tensor_trans_insts.instructions
    if instructions is None or len(instructions) != 2:
      return

    instr_0, instr_1 = instructions
    params_0 = instr_0.parameters
    params_1 = instr_1.parameters
    producer_op_index = instr_0.producer
    if (
        not isinstance(params_0, qtyping.UniformQuantParams)
        or not isinstance(params_1, qtyping.UniformQuantParams)
        or not self._is_valid_quantize_requantize_pair(instr_0, instr_1)
        or not self._are_quant_params_compatible(params_0, params_1)
        # To avoid fusion when subgraph inputs connected to the main subgraph
        # (e.g. while_body), we skip all tensors with no producer.
        or producer_op_index == -1
        # Can't apply fusion to tensors with a constrained producer since that
        # will break the constraint.
        or self._is_op_constrained(
            tensor_trans_insts.subgraph_id, producer_op_index
        )
    ):
      return

    # Fuse the quantize and requantize.
    instr_0.parameters = dataclasses.replace(
        params_0, scale=params_1.scale, zero_point=params_1.zero_point
    )
    # Remove the requantize instruction.
    instructions.pop(1)

  def _quant_params_to_transformation_insts(
      self,
      param: qtyping.TensorTransformationParams,
  ) -> qtyping.TensorTransformationInsts:
    """Convert single tensor quant params to transformation instructions.

    Args:
      param: Quantization parameters of a tensor in the graph.

    Returns:
      Transformations to be applied to the given tensor.
    """
    # Setup the structure.
    tensor_info = self._tensor_name_to_graph_info[param.tensor_name]
    tensor_trans_insts = qtyping.TensorTransformationInsts(
        param.tensor_name, tensor_info.subgraph_id, []
    )

    # Add all producer rules.
    transformations = []
    if param.producer:
      for transformation in param.producer.transformations:
        transformations.append(
            qtyping.TransformationInst(
                transformation,
                tensor_info.tensor_id,
                tensor_info.producer,
                tensor_info.consumers,
                param.producer.parameters,
            )
        )

    # Horizontal optimization.
    consumer_group = self._group_consumer_transformations(param)
    # At this point, starting from index 1 of consumer_group, we're having sets
    # that represent transformations that can be grouped together.
    transformations_available_for_vertical_optimization = (
        self._produce_transformation_for_vertical_opt(consumer_group, param)
    )
    other_consumer_transformations = (
        self._produce_consumer_transformations_unavailable_for_vertical_opt(
            consumer_group, param
        )
    )
    # Apply vertical optimization.
    last_producer_rule_idx = len(transformations) - 1
    if last_producer_rule_idx >= 0:
      transformations += self._apply_vertical_optimization(
          transformations.pop(),
          transformations_available_for_vertical_optimization,
      )
    else:
      transformations += transformations_available_for_vertical_optimization
    # Adding other consumers rules.
    transformations += other_consumer_transformations
    tensor_trans_insts.instructions = transformations

    # Now, when all optimizations are done, we can remove the last tensor
    # duplication instruction, so the original tensor can be reused.
    self._remove_last_tensor_duplication(tensor_trans_insts)
    # With the tensor duplication instructions finalized, we can remove
    # unnecessary buffer duplications applied to the same duplicated tensors.
    # This is not a part of a vertical optimization because vertical
    # optimization only works between producers & consumers, and this is between
    # the consumer only. Also this can't be done during the params generation
    # because removing last tensor duplication has to happen first.
    self._remove_unnecessary_buffer_duplication(tensor_trans_insts)

    # Check the generated transformation instructions are valid, the function
    # will raise an error if the instructions are not valid.
    self._check_tensor_transformation_instructions_valid(tensor_trans_insts)

    # Remove unnecessary [QUANTIZE_TENSOR, ADD_QUANTIZE] pairs for tensors with
    # providers without same as input/output scale constraints.
    self._eliminate_requantization_for_nonconstrained_provider(
        tensor_trans_insts
    )

    return tensor_trans_insts

  def _split_instructions_by_tensor_duplication(
      self,
      instructions: qtyping.TensorTransformationInsts,
  ) -> list[list[qtyping.TransformationInst]]:
    """Split the instructions into subsets by tensor duplication.

    Splits the instructions into subsets based on which tensor (original or one
    of duplicated ones) they will be applied to.

    The first subset is for the original tensor. The following subsets are for
    the duplicated tensors. The order of instructions in each subset is
    preserved.

    Enforced constraints for each duplicated tensor's instructions subset:
    1. The first instruction must be a `DUPLICATE_TENSOR` one.
    2. No other `DUPLICATE_TENSOR` instructions can be present.

    For the following instructions:
      [
          (transformation=DUPLICATE_TENSOR, consumers=[1, 2, 3]),
          (transformation=DUPLICATE_TENSOR, consumers=[4]),
          (transformation=T1, consumers=[1, 2]),
          (transformation=T2, consumers=[3]),
          (transformation=T3, consumers=[4]),
          (transformation=T4, consumers=[5])
      ]

    `instruction_subsets` will be:
      [
          [(transformation=T4, consumers=[5])],
          [
              (transformation=DUPLICATE_TENSOR, consumers=[1, 2, 3]),
              (transformation=T1, consumers=[1, 2]),
              (transformation=T2, consumers=[3])
          ],
          [
              (transformation=DUPLICATE_TENSOR, consumers=[4]),
              (transformation=T3, consumers=[4])
          ]
      ],

    Args:
      instructions: Transformation instructions for a tensor.

    Returns:
      A list of subsets of transformation instructions, where the first subset
      is for the original tensor, and the following subsets are for the
      duplicated tensors.

    Raises:
      ValueError: If DUPLICATE_TENSOR is found and it's not the first
      transformation for its consumers.
    """
    original_tensor_subset_idx = 0
    instruction_subsets = [[]]
    consumer_to_subset_idx = {}
    for instruction in instructions.instructions:
      if instruction.transformation == _QuantTransformation.DUPLICATE_TENSOR:
        instruction_subsets.append([instruction])
        subset_idx = len(instruction_subsets) - 1
        for consumer in instruction.consumers:
          if consumer in consumer_to_subset_idx:
            raise ValueError(
                f"Tensor {instructions.tensor_name} : duplicate tensor should"
                " be the first instruction for its consumers."
            )
          else:
            consumer_to_subset_idx[consumer] = subset_idx
      else:
        first_consumer = instruction.consumers[0]
        if first_consumer not in consumer_to_subset_idx:
          consumer_to_subset_idx[first_consumer] = original_tensor_subset_idx
        subset_idx = consumer_to_subset_idx[first_consumer]
        instruction_subsets[subset_idx].append(instruction)

    return instruction_subsets

  def _check_subset_of_tensor_transformation_instructions_valid(
      self,
      instructions: Optional[list[qtyping.TransformationInst]],
      tensor_name: str,
  ):
    """Check if a subset of tensor transformation instructions is valid.

    Args:
      instructions: A subset of transformation instructions for a tensor.
      tensor_name: The name of the tensor.

    Raises:
      ValueError: If the subset of instructions are not valid.
    """
    is_tensor_unquantized = False
    is_tensor_quantized = False
    for instruction in instructions:
      transform_type = instruction.transformation
      if transform_type == qtyping.QuantTransformation.NO_QUANTIZE:
        is_tensor_unquantized = True
      elif (
          transform_type == qtyping.QuantTransformation.QUANTIZE_TENSOR
          or transform_type == qtyping.QuantTransformation.ADD_DEQUANTIZE
      ):
        is_tensor_quantized = True
    if is_tensor_unquantized and is_tensor_quantized:
      raise ValueError(
          "Tensor %s can not be both quantized and unquantized" % tensor_name
      )

  def _check_tensor_transformation_instructions_valid(
      self,
      instructions: qtyping.TensorTransformationInsts,
  ):
    """Check if the tensor transformation instructions are valid.

    Args:
      instructions: Transformation instructions for a tensor.

    Raises:
      ValueError: If the instructions are not valid.
    """
    # Split the instructions into subsets based on which tensor (original or one
    # of duplicated ones) they will be applied to.
    instruction_subsets = self._split_instructions_by_tensor_duplication(
        instructions
    )
    # Check that each subset of instructions is valid.
    for instruction_subset in instruction_subsets:
      self._check_subset_of_tensor_transformation_instructions_valid(
          instruction_subset,
          instructions.tensor_name,
      )

  def quant_params_to_transformation_insts(
      self,
      params: dict[str, qtyping.TensorTransformationParams],
      flatbuffer_model: Optional[schema_py_generated.ModelT] = None,
  ) -> dict[str, qtyping.TensorTransformationInsts]:
    """Converts quantization params to transformation instructions.

    Args:
      params: quantization parameters generated by params_generator. The data
        type is designed to be the same as the output of
        generate_quantization_parameters.
      flatbuffer_model: the flatbuffer model to be quantized.

    Returns:
      a dictionary with tensor name as key and transformation instructions as
      value
    """
    if flatbuffer_model is not None:
      self.flatbuffer_model = flatbuffer_model
      self._create_tensor_name_to_graph_info_map()

    insts = {}
    for tensor_name in params:
      insts[tensor_name] = self._quant_params_to_transformation_insts(
          params[tensor_name]
      )
    return insts



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A command-line tool to quantize TFLite models using a quantization recipe."""

import argparse
from collections.abc import Sequence
import pathlib
import sys

import os
from ai_edge_quantizer import quantizer


def quantize_model(
    model_file: str,
    recipe_file: str,
    output_dir: str,
    overwrite_outputs: bool = False,
):
  """Quantizes a TFLite model using a recipe.

  Args:
    model_file: Path to the .tflite file to be quantized.
    recipe_file: Path to the .json file with the quantization recipe.
    output_dir: Path to the directory to save the quantized model(s).
    overwrite_outputs: Output files overwrite exisiting files without requiring
      user input.
  """

  model_basename = pathlib.Path(model_file).stem
  recipe_basename = pathlib.Path(recipe_file).stem
  output_filename = f"{model_basename}_{recipe_basename}.tflite"

  if not os.path.exists(output_dir):
    os.makedirs(output_dir)

  output_file_path = str(pathlib.Path(output_dir) / output_filename)

  overwrite = overwrite_outputs
  if os.path.exists(output_file_path) and not overwrite:
    overwrite_input = input(
        f"Output file {output_file_path} already exists. Overwrite? (y/N): "
    )
    if overwrite_input.lower() != "y":
      print("Aborting.")
      return
    overwrite = True

  qt = quantizer.Quantizer(model_file)
  qt.load_quantization_recipe(recipe_file)
  qt.quantize().export_model(output_file_path, overwrite=overwrite)


def parse_args(args: Sequence[str]) -> argparse.Namespace:
  """Parses command-line arguments.

  Args:
    args: A list of strings to parse. If None, sys.argv is used.

  Returns:
    An argparse.Namespace containing the parsed arguments,
  """
  parser = argparse.ArgumentParser(
      description="Quantize TFLite models using a quantization recipe.",
  )
  parser.add_argument(
      "--model_file",
      required=True,
      help="Path to the .tflite file to be quantized.",
  )
  parser.add_argument(
      "--recipe_file",
      required=True,
      help="Path to the .json file with the quantization recipe.",
  )
  parser.add_argument(
      "--output_dir",
      required=True,
      help="Path to the directory in which to save the quantized model(s).",
  )
  parser.add_argument(
      "--overwrite_outputs",
      action="store_true",
      help="Overwrite exisiting output files without requesting user input.",
  )
  return parser.parse_args(args[1:])


def main(parsed_args: argparse.Namespace):
  quantize_model(
      model_file=parsed_args.model_file,
      recipe_file=parsed_args.recipe_file,
      output_dir=parsed_args.output_dir,
      overwrite_outputs=parsed_args.overwrite_outputs,
  )


if __name__ == "__main__":
  main(parse_args(sys.argv))



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""The Python API for Algorithm Manager of Quantizer."""

from collections.abc import Callable
import dataclasses
import functools
from typing import Any, Optional
from ai_edge_quantizer import qtyping


@dataclasses.dataclass
class QuantizedOperationInfo:
  """Stores all quantization functions for a given op."""

  tfl_op_key: qtyping.TFLOperationName
  init_qsv_func: Callable[..., Any]
  calibration_func: Callable[..., Any]
  materialize_func: Callable[..., Any]


@dataclasses.dataclass
class QuantizationAlgorithmInfo:
  quantization_algorithm: str
  quantized_ops: dict[qtyping.TFLOperationName, QuantizedOperationInfo]


class AlgorithmManagerApi:
  """Quantizer API client to manage quantization configs and functions."""

  def __init__(self):
    self._algorithm_registry = dict()
    # Check if an op quantization config is supported for a given algorithm.
    self._config_check_registry = dict()
    # Policy to check if an op quantization config is supported for a given
    # algorithm.
    self._config_check_policy_registry = dict()

  def register_op_quant_config_validation_func(
      self,
      algorithm_key: str,
      config_check_func: Callable[..., Any],
  ):
    """Register functions to check if an op quantization config is supported."""
    self._config_check_registry[algorithm_key] = config_check_func

  def register_quantized_op(
      self,
      algorithm_key: str,
      tfl_op_name: qtyping.TFLOperationName,
      init_qsv_func: Callable[..., Any],
      calibration_func: Callable[..., Any],
      materialize_func: Callable[..., Any],
  ):
    """Register functions to support a quantization operation.

    This function registers the relevant information to support the quantized
    version of given tfl_operation, for the algorithm specified by
    quantization_algorithm.

    Args:
      algorithm_key: Quantization algorithm keyword for which the quantized
        operation is for.
      tfl_op_name: TFLite op name.
      init_qsv_func: QSV init function to be called.
      calibration_func: Quantized operation to be called during calibration.
      materialize_func: Quantized operation to be called during materialization.
    """
    quantized_algorithm_info = self._algorithm_registry.setdefault(
        algorithm_key, QuantizationAlgorithmInfo(algorithm_key, dict())
    )

    quantized_algorithm_info.quantized_ops[tfl_op_name] = (
        QuantizedOperationInfo(
            tfl_op_name,
            init_qsv_func,
            calibration_func,
            materialize_func,
        )
    )

  def is_op_registered(
      self,
      quantization_algorithm: str,
      tfl_op_name: qtyping.TFLOperationName,
  ) -> bool:
    """Check if the given key for quantization is valid.

    Args:
      quantization_algorithm: Target quantization algorithm.
      tfl_op_name: TFL operation name.

    Returns:
      True if the given op is registered for the given algorithm, false
      otherwise.
    """
    if not self.is_algorithm_registered(quantization_algorithm):
      return False

    return (
        tfl_op_name
        in self._algorithm_registry[quantization_algorithm].quantized_ops
    )

  def is_algorithm_registered(self, quantization_algorithm: str) -> bool:
    """Check if the given algorithm is registered.

    Args:
      quantization_algorithm: Target quantization algorithm.

    Returns:
      True if the given algorithm is registered, false otherwise.
    """
    return quantization_algorithm in self._algorithm_registry

  def check_op_quantization_config(
      self,
      quantization_algorithm: str,
      tfl_op_name: qtyping.TFLOperationName,
      op_quantization_config: qtyping.OpQuantizationConfig,
  ) -> None:
    """Checks if the given op quantization config is valid.

    Args:
      quantization_algorithm: Target quantization algorithm.
      tfl_op_name: TFL operation name.
      op_quantization_config: Op quantization config to be checked.

    Raises:
      ValueError if the given op is not registered for the given algorithm, or
      the given algorithm is not registered.
    """
    if op_quantization_config.skip_checks:
      return
    if not self.is_op_registered(quantization_algorithm, tfl_op_name):
      raise ValueError(
          f"Unsupported operation {tfl_op_name} for Algorithm:"
          f" {quantization_algorithm}."
      )
    if quantization_algorithm not in self._config_check_registry:
      raise ValueError(
          f"Config checking function for  algorithm {quantization_algorithm} is"
          " not registered. Please use"
          " `register_op_quant_config_validation_func` to register the"
          " validation function."
      )
    self._config_check_registry[quantization_algorithm](
        tfl_op_name,
        op_quantization_config,
        self._config_check_policy_registry[quantization_algorithm],
    )

  def get_supported_ops(self, alg_key: str) -> list[qtyping.TFLOperationName]:
    """Returns the list of supported ops for the given algorithm.

    Args:
      alg_key: Algorithm key.

    Returns:
      The list of supported JAX operations.

    Raises:
      ValueError if the alg_key is not registered.
    """
    if alg_key not in self._algorithm_registry:
      raise ValueError(f"Unregistered algorithm: {alg_key}")

    return list(self._algorithm_registry[alg_key].quantized_ops.keys())

  def get_quantization_func(
      self,
      algorithm_key: str,
      tfl_op_name: qtyping.TFLOperationName,
      quantize_mode: qtyping.QuantizeMode,
  ) -> Callable[..., Any]:
    """Gets the quantization function.

    Args:
      algorithm_key: Target quantization algorithm key (e.g.,
        AlgorithmName.MIN_MAX_UNIFORM_QUANT).
      tfl_op_name: TFLite op name.
      quantize_mode: Quantization mode to be used.

    Returns:
      A quantized operation (function) corresponds to the requested algorithm
      for the TFL op.
    """
    if not self.is_op_registered(algorithm_key, tfl_op_name):
      raise ValueError(
          f"Unsupported operation {tfl_op_name} for Algorithm: {algorithm_key}."
          f" Supported ops for algorithm {algorithm_key}:"
          f" {self.get_supported_ops(algorithm_key)}"
      )

    quantized_algorithm_info = self._algorithm_registry[algorithm_key]
    quantized_op_info = quantized_algorithm_info.quantized_ops
    quantized_func = self._get_target_func(
        quantized_op_info, tfl_op_name, quantize_mode
    )
    if quantized_func is None:
      raise ValueError(
          "Cannot retrieve appropriate quantization function for"
          f" {tfl_op_name} for algorithm {algorithm_key} under quantization"
          f" mode {quantize_mode}. Check if the op is registed in"
          " algorithm_manager."
      )

    return quantized_func

  def get_init_qsv_func(
      self,
      algorithm_key: str,
      tfl_op_name: qtyping.TFLOperationName,
  ) -> functools.partial:
    """Gets the initial Quantization Statistics Variable function for a given op.

    Args:
      algorithm_key: Quantization algorithm to search.
      tfl_op_name: Target TFL operation.

    Returns:
      A function for qsv initialization.
    """

    if not self.is_op_registered(algorithm_key, tfl_op_name):
      raise ValueError(
          f"Unsupported operation {tfl_op_name} for Algorithm: {algorithm_key}."
          f" Supported ops for algorithm {algorithm_key}:"
          f" {self.get_supported_ops(algorithm_key)}"
      )
    quantized_algorithm_info = self._algorithm_registry[algorithm_key]
    quantized_op_info = quantized_algorithm_info.quantized_ops

    return quantized_op_info[tfl_op_name].init_qsv_func

  def _get_target_func(
      self,
      quantized_op_info,
      tfl_op_name: qtyping.TFLOperationName,
      quantize_mode: qtyping.QuantizeMode,
  ):
    """Gets the function corresponding to the given JAX quantization phase and op."""
    if quantize_mode == qtyping.QuantizeMode.CALIBRATE:
      return quantized_op_info[tfl_op_name].calibration_func
    elif quantize_mode == qtyping.QuantizeMode.MATERIALIZE:
      return quantized_op_info[tfl_op_name].materialize_func
    return None

  # TODO: b/53780772 - Merge this function with
  # register_op_quant_config_validation_func after full transition to new check
  # mechanism.
  def register_config_check_policy(
      self,
      algorithm_key: str,
      config_check_policy: Optional[qtyping.ConfigCheckPolicyDict],
  ):
    """Registers a policy to check the op quantization config."""
    self._config_check_policy_registry[algorithm_key] = config_check_policy



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test for validation_utils."""
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer.utils import validation_utils


class ValidationUtilTest(absltest.TestCase):

  def test_preprocess_same_size_arrays_nan(self):
    data1 = [1, 2, np.nan]
    data2 = [1, 2, 3]
    processed_data1, processed_data2 = (
        validation_utils._preprocess_same_size_arrays(data1, data2)
    )
    self.assertAlmostEqual(processed_data1[2], 1e-9)
    self.assertListEqual(list(processed_data2), data2)

  def test_preprocess_same_size_arrays_inf(self):
    data1 = [1, 2, -np.inf]
    data2 = [1, 2, np.inf]
    processed_data1, processed_data2 = (
        validation_utils._preprocess_same_size_arrays(data1, data2)
    )
    self.assertAlmostEqual(processed_data1[2], -1e9)
    self.assertAlmostEqual(processed_data2[2], 1e9)

  def test_preprocess_same_size_arrays_fails_with_incorrect_shape(self):
    data1 = [1, 2, 3]
    data2 = [1, 2]
    self.assertRaises(
        ValueError, validation_utils._preprocess_same_size_arrays, data1, data2
    )

  def test_mean_squared_difference(self):
    data1 = [1, 2, 3]
    data2 = [1, 2, 3]
    result = validation_utils.mean_squared_difference(data1, data2)
    self.assertEqual(result, 0)

  def test_mean_squared_difference_multidim(self):
    data1 = [[1, 2], [4, 5]]
    data2 = [[1, 3], [2, 2]]
    result = validation_utils.mean_squared_difference(data1, data2)
    self.assertAlmostEqual(result, 3.5)

  def test_mean_squared_difference_0d(self):
    data1 = []
    data2 = []
    result = validation_utils.mean_squared_difference(data1, data2)
    self.assertEqual(result, 0)

  def test_median_diff_ratio(self):
    data1 = [1, 2, 3]
    data2 = [1, 2, 3]
    result = validation_utils.median_diff_ratio(data1, data2)
    self.assertEqual(result, 0)

  def test_median_diff_ratio_multidim(self):
    data1 = [[1, 2], [4, 5]]
    data2 = [[1, 3], [2, 2]]
    result = validation_utils.median_diff_ratio(data1, data2)
    self.assertAlmostEqual(result, 0.6666664)

  def test_median_diff_ratio_0d(self):
    data1 = []
    data2 = []
    result = validation_utils.median_diff_ratio(data1, data2)
    self.assertEqual(result, 0)

  def test_cosine_similarity(self):
    data1 = [1, 2, 3]
    data2 = [1, 2, 3]
    result = validation_utils.cosine_similarity(data1, data2)
    self.assertAlmostEqual(result, 1.0, 6)

  def test_cosine_similarity_perpendicular(self):
    data1 = [1, 0, 0]
    data2 = [0, 1, 0]
    result = validation_utils.cosine_similarity(data1, data2)
    self.assertAlmostEqual(result, 0.0, 6)

  def test_cosine_similarity_multidim(self):
    data1 = [[1, 2], [4, 5]]
    data2 = [[1, 3], [2, 2]]
    result = validation_utils.cosine_similarity(data1, data2)
    self.assertAlmostEqual(result, 0.86881, 6)

  def test_cosine_similarity_0d(self):
    data1 = []
    data2 = []
    result = validation_utils.cosine_similarity(data1, data2)
    self.assertEqual(result, 0)

  def test_kl_divergence(self):
    data1 = [0.5, 0.5]
    data2 = [0.1, 0.9]
    result = validation_utils.kl_divergence(data1, data2)
    self.assertAlmostEqual(result, 0.36808, 4)

  def test_kl_divergence_zero_in_q(self):
    data1 = [0, 1]
    data2 = [1, 0]
    result = validation_utils.kl_divergence(data1, data2)
    self.assertAlmostEqual(result, 20.7232658, 4)

  def test_kl_divergence_negative_values(self):
    data1 = [-1, 1]
    data2 = [1, -1]
    result = validation_utils.kl_divergence(data1, data2)
    self.assertAlmostEqual(result, 20.7232658, 4)

  def test_kl_divergence_0d(self):
    data1 = []
    data2 = []
    result = validation_utils.kl_divergence(data1, data2)
    self.assertEqual(result, 0)

  def test_get_validation_func_kl_divergence(self):
    func = validation_utils.get_validation_func("kl_divergence")
    self.assertEqual(func, validation_utils.kl_divergence)

  def test_signal_to_noise_ratio_0d(self):
    data1 = []
    data2 = []
    result = validation_utils.signal_to_noise_ratio(data1, data2)
    self.assertEqual(result, 0)

  def test_signal_to_noise_ratio_identical(self):
    data1 = [1, 2, 3]
    data2 = [1, 2, 3]
    result = validation_utils.signal_to_noise_ratio(data1, data2)
    self.assertGreater(result, 1e8)  # mse=0, so snr should be large

  def test_signal_to_noise_ratio_with_noise(self):
    data1 = [2, 3, 4]
    data2 = [1, 2, 3]
    result = validation_utils.signal_to_noise_ratio(data1, data2)
    self.assertAlmostEqual(result, 14 / 3, places=5)

  def test_signal_to_noise_ratio_simple(self):
    data1 = [1, 1]
    data2 = [1, 0]
    result = validation_utils.signal_to_noise_ratio(data1, data2)
    self.assertAlmostEqual(result, 1.0, places=5)

  def test_get_validation_func_snr(self):
    func = validation_utils.get_validation_func("snr")
    self.assertEqual(func, validation_utils.signal_to_noise_ratio)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Common utils for uniform quantization algorithms."""

from collections.abc import MutableMapping, Sequence
import dataclasses
import enum
from typing import Any, Optional
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

_TFLOpName = qtyping.TFLOperationName
_QuantTransformation = qtyping.QuantTransformation
_IntType = uniform_quantize_tensor.IntType
TensorQuantParamsCacheKey = tuple[Any, qtyping.TensorQuantizationConfig]


_DRQ_OR_WEIGHT_ONLY_OPS = frozenset([
    _TFLOpName.FULLY_CONNECTED,
    _TFLOpName.CONV_2D,
    _TFLOpName.BATCH_MATMUL,
    _TFLOpName.EMBEDDING_LOOKUP,
    _TFLOpName.DEPTHWISE_CONV_2D,
    _TFLOpName.CONV_2D_TRANSPOSE,
])

_SUPPORTED_SUBCHANNEL_OPS = frozenset([
    _TFLOpName.FULLY_CONNECTED,
    _TFLOpName.EMBEDDING_LOOKUP,
])


class TensorQuantParamsCache:
  """Cache of `UniformQuantParams|NonLinearQuantParams` objects.

  Cache of `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of
  the buffer ID and the `TensorQuantizationConfig` used to compute it.
  """

  def __init__(self):
    self._cache: MutableMapping[
        TensorQuantParamsCacheKey,
        qtyping.UniformQuantParams | qtyping.NonLinearQuantParams,
    ] = {}

  def lookup(
      self, buffer_id: int, quant_config: qtyping.TensorQuantizationConfig
  ) -> qtyping.UniformQuantParams | qtyping.NonLinearQuantParams | None:
    cache_key: TensorQuantParamsCacheKey = (
        buffer_id,
        quant_config,
    )
    return self._cache.get(cache_key, None)

  def insert(
      self,
      buffer_id: int,
      quant_config: qtyping.TensorQuantizationConfig,
      quant_params: qtyping.UniformQuantParams | qtyping.NonLinearQuantParams,
  ) -> qtyping.UniformQuantParams | qtyping.NonLinearQuantParams:
    self._cache[(buffer_id, quant_config)] = quant_params
    return quant_params


def check_subchannel_config(
    op_name: _TFLOpName, op_quant_config: qtyping.OpQuantizationConfig
):
  """Checks the op quantization config for subchannel quantization."""
  if (
      op_quant_config.weight_tensor_config is not None
      and uniform_quantize_tensor.is_blockwise(
          op_quant_config.weight_tensor_config.granularity
      )
  ):
    if op_name not in _SUPPORTED_SUBCHANNEL_OPS:
      raise ValueError(f"Unsupported op for blockwise quantization: {op_name}.")
    if op_quant_config.activation_tensor_config is not None:
      raise ValueError(
          "Blockwise quantization does not support activation tensor"
          " quantization."
      )
    if not op_quant_config.weight_tensor_config.symmetric:
      raise ValueError(
          "Blockwise quantization does not support for asymmetric weight"
          " quantization."
      )


def check_if_valid_op_config(
    op_name: _TFLOpName,
    op_quant_config: qtyping.OpQuantizationConfig,
    config_check_policy: qtyping.ConfigCheckPolicyDict,
) -> None:
  """Check if the op quantization config is valid.

  Args:
    op_name: The name of the op.
    op_quant_config: The quantization config for the op.
    config_check_policy: The policy to check the op quantization config.

  Raises:
    ValueError: If the op quantization config is not valid.
  """
  check_passed = False
  error_msg = ""
  # Check if find op_config in policy config_check_policy.
  if config_check_policy is None:
    error_msg = "No policy was specified at all."
  elif op_name not in config_check_policy.keys():
    error_msg = (
        f"No policy was specified for op: {op_name} with config:"
        f" {op_quant_config}."
    )
  # The config_check_policy contains all possible valid configs, except for
  # variations in the min_weight_elements field (it's set to 0 for all of them).
  # min_weight_elements has to be ignored during policy check here because it
  # can be any non-negative integer, which means we can't list all possible
  # values in the policy.
  elif (
      dataclasses.replace(op_quant_config, min_weight_elements=0)
      not in config_check_policy[op_name]
  ):
    error_msg = (
        f"Quantization config for op: {op_name} with config:"
        f" {op_quant_config} was not found in the policy."
    )
  else:
    check_passed = True

  if not check_passed:
    raise ValueError(
        f"Unsupported op for {op_quant_config.compute_precision}: {op_name}."
        f" Error: {error_msg}"
    )


class OpQuantConstraint(enum.Enum):
  """Quantization constraint for an op."""

  NO_CONSTRAIN = 0
  # All tensors in the op have the same scale as the input tensor
  # e.g., transpose/reshape/split.
  SAME_AS_INPUT_SCALE = 1
  # All tensors in the op have the same scale as the output tensor.
  # e.g., concatenate
  SAME_AS_OUTPUT_SCALE = 2


def _get_tensor_transformation_params_wrapper(
    tensor: Any,
    is_inbounding_tensor: bool,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
    quant_params=None,
) -> qtyping.TensorTransformationParams:
  """Util to get tensor transformation params.

  Args:
    tensor: Tensor to be quantized.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    quant_params: Quantization parameters for the tensor.

  Returns:
    Transformation parameters for the tensor.

  Raises:
    ValueError: If the tensor is not found in tensor_name_to_qsv.
  """
  tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
  tensor_data = tfl_flatbuffer_utils.get_tensor_data(tensor, graph_info.buffers)
  tensor_quant_config = op_info.op_quant_config.activation_tensor_config
  is_constant = tensor_data is not None
  # Use weight configuration if it is supported.
  if is_constant and op_info.op_name in _DRQ_OR_WEIGHT_ONLY_OPS:
    tensor_quant_config = op_info.op_quant_config.weight_tensor_config
  # Get quant params.
  if quant_params is None and tensor_quant_config is not None:
    # Is there possibly a cached version of this value?
    if not is_constant or not (
        quant_params := tensor_quant_params_cache.lookup(
            tensor.buffer, tensor_quant_config
        )
    ):
      quant_params = get_tensor_quant_params_fn(
          op_info,
          tensor_quant_config,
          tensor_data,
          tensor_name_to_qsv.get(tensor_name),
      )
      # Update the cache if we have a key.
      if is_constant:
        tensor_quant_params_cache.insert(
            tensor.buffer, tensor_quant_config, quant_params
        )
  return get_tensor_transformation_params(
      tensor_name,
      op_info,
      is_inbounding_tensor,
      quant_params,
      is_constant,
  )


def _materialize_op_tensors(
    op_tensor_params: list[qtyping.TensorTransformationParams],
    op_tensors: Sequence[Any],
    is_inbounding_tensor: bool,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
    quant_params=None,
) -> None:
  """Util to materialize op tensors. Appends the results to op_tensor_params.

  Args:
    op_tensor_params: Tensor transformation parameters for the op. Will be
      modified to include new tensor parameters.
    op_tensors: Tensors associated with the op.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    quant_params: Quantization parameters for the tensor.
  """
  for tensor in op_tensors:
    tensor_params = _get_tensor_transformation_params_wrapper(
        tensor,
        is_inbounding_tensor,
        op_info,
        graph_info,
        tensor_name_to_qsv,
        get_tensor_quant_params_fn=get_tensor_quant_params_fn,
        quant_params=quant_params,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )
    op_tensor_params.append(tensor_params)


def _get_single_tensor_params(
    tensors: Sequence[Any],
    is_inbounding_tensor: bool,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> qtyping.TensorTransformationParams:
  """Util to get single tensor params.

  Args:
    tensors: A list of a single tensor.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Transformation parameters for the tensor.

  Raises:
    ValueError: If the tensor list is not of size 1.
  """
  if len(tensors) != 1:
    raise ValueError(
        "Trying to get a single tensor params with a list of multiple tensor"
        f" with size {len(tensors)}."
    )
  return _get_tensor_transformation_params_wrapper(
      tensors[0],
      is_inbounding_tensor,
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def _materialize_tensors_with_quantized_data_update(
    op_tensor_params: list[qtyping.TensorTransformationParams],
    tensors: Sequence[Any],
    quant_params: Optional[qtyping.UniformQuantParams],
    is_inbounding_tensor: bool,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> None:
  """Materialize a list of tensors with `quantized_data` updated when needed.

  Args:
    op_tensor_params: Tensor transformation parameters for the op. Will be
      modified to include new tensor parameters.
    tensors: Tensors to be materialized.
    quant_params: The quantization parameters to be used for materialization.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
  """
  if quant_params is not None and quant_params.quantized_data is not None:
    quant_params = dataclasses.replace(quant_params, quantized_data=None)

  for tensor in tensors:
    tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        tensor, graph_info.buffers
    )
    if quant_params is None or tensor_data is None:
      tensor_quant_params = quant_params
    else:
      # Constant tensors require updating `quantized_data`.
      quantized_data = uniform_quantize_tensor.uniform_quantize(
          tensor_data, quant_params
      )
      tensor_quant_params = dataclasses.replace(
          quant_params,
          quantized_data=quantized_data,
      )
    _materialize_op_tensors(
        op_tensor_params,
        [tensor],
        is_inbounding_tensor=is_inbounding_tensor,
        op_info=op_info,
        graph_info=graph_info,
        tensor_name_to_qsv=tensor_name_to_qsv,
        get_tensor_quant_params_fn=get_tensor_quant_params_fn,
        quant_params=tensor_quant_params,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )


def _materialize_standard_op_with_same_as_input_scale(
    input_tensors: Sequence[Any],
    output_tensors: Sequence[Any],
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in an op with same as input scale constraint.

  Args:
    input_tensors: Input tensors for the op.
    output_tensors: Output tensors for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  op_tensor_params = []
  # Must be a single input to avoid ambiguity.
  input_tensor_params = _get_single_tensor_params(
      input_tensors,
      is_inbounding_tensor=True,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )
  op_tensor_params.append(input_tensor_params)
  # Use input quantization params for all output tensors.
  input_quant_params = input_tensor_params.consumers[0].parameters
  if not isinstance(input_quant_params, qtyping.UniformQuantParams):
    raise ValueError(
        "_materialize_standard_op_with_same_as_input_scale only supports"
        f" UniformQuantParams. For tensor {input_tensor_params.tensor_name},"
        f" got {type(input_quant_params)}"
    )
  _materialize_tensors_with_quantized_data_update(
      op_tensor_params,
      output_tensors,
      input_quant_params,
      is_inbounding_tensor=False,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )

  # Change output qsv to be the same as input qsv. This is safe since TFL
  # subgraph is acyclic.
  input_tensor_qsv = tensor_name_to_qsv.get(
      input_tensor_params.tensor_name, None
  )
  if input_tensor_qsv is None:
    input_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        input_tensors[0], graph_info.buffers
    )
    # If the input tensor is a constant tensor without qsv, compute qsv from
    # its quant params.
    if input_tensor_data is None:
      # If the only input to an op that needs to match input to
      # output has no qsv and is not a constant tensor, then this is an error.
      raise ValueError(
          "Input tensor qsv is None for tensor"
          f" {input_tensor_params.tensor_name}."
      )
    min_val, max_val = _get_min_max_from_quant_params(input_quant_params)
    input_tensor_qsv = {"min": min_val, "max": max_val}
  for output_tensor in output_tensors:
    tensor_name_to_qsv[
        tfl_flatbuffer_utils.get_tensor_name(output_tensor)
    ] = input_tensor_qsv

  return op_tensor_params


def _materialize_standard_op_with_same_as_output_scale(
    input_tensors: Sequence[Any],
    output_tensors: Sequence[Any],
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in an op with same as output scale constraint.

  Args:
    input_tensors: Input tensors for the op.
    output_tensors: Output tensors for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  op_tensor_params = []
  # Must be a single output to avoid ambiguity.
  output_tensor_params = _get_single_tensor_params(
      output_tensors,
      is_inbounding_tensor=False,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )
  # Use output quantization params for all input tensors.
  if output_tensor_params.producer is None:
    output_quant_params = None
  else:
    output_quant_params = output_tensor_params.producer.parameters
    if not isinstance(output_quant_params, qtyping.UniformQuantParams):
      raise ValueError(
          "_materialize_standard_op_with_same_as_output_scale only supports"
          f" UniformQuantParams. For tensor {output_tensor_params.tensor_name},"
          f" got {type(output_quant_params)}"
      )
  _materialize_tensors_with_quantized_data_update(
      op_tensor_params,
      input_tensors,
      output_quant_params,
      is_inbounding_tensor=True,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )

  op_tensor_params.append(output_tensor_params)

  return op_tensor_params


def _materialize_standard_op_no_constraint(
    input_tensors: Sequence[Any],
    output_tensors: Sequence[Any],
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in an op with no constraint.

  Args:
    input_tensors: Input tensors for the op.
    output_tensors: Output tensors for the op.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  op_tensor_params = []
  _materialize_op_tensors(
      op_tensor_params,
      input_tensors,
      is_inbounding_tensor=True,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )
  _materialize_op_tensors(
      op_tensor_params,
      output_tensors,
      is_inbounding_tensor=False,
      op_info=op_info,
      graph_info=graph_info,
      tensor_name_to_qsv=tensor_name_to_qsv,
      get_tensor_quant_params_fn=get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )

  return op_tensor_params


def _split_tensors_by_indices(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    indices: Optional[Sequence[int]],
    is_inbounding_tensor: bool,
) -> tuple[list[Any], list[Any], list[int]]:
  """Split tensors into two lists and return indices with -1 values removed.

  This function splits the tensors into two lists based on the provided indices.
    * The first list contains tensors with indices in the provided indices list.
    * The second list contains all remaining tensors.

  Additionally, the function filters out any tensors with the index -1
  (indicating non-existing bias in FC and cov ops) and returns a new list of
  indices excluding these values.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    indices: Indices of tensors to use for split.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor for the op.

  Returns:
    A tuple containing:
      * A list of tensors with indices in the provided list.
      * A list of tensors with indices not in the provided list.
      * A new list of indices with -1 values removed.
  """
  indices = indices or []
  updated_indices = []
  updated_index = 0
  selected_tensors, others = [], []
  tensors = op_info.op.inputs if is_inbounding_tensor else op_info.op.outputs
  for i, tensor_index in enumerate(tensors):
    # Ignore non-existing tensors.
    if tensor_index == -1:
      continue
    if i in indices:
      updated_indices.append(updated_index)
      selected_tensors.append(graph_info.subgraph_tensors[tensor_index])
    else:
      others.append(graph_info.subgraph_tensors[tensor_index])
    updated_index += 1

  return selected_tensors, others, updated_indices


def _materialize_ignored_tensors(
    tensors: Sequence[Any],
    op_info: qtyping.OpInfo,
    is_inbounding_tensor: bool,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize ignored tensors.

  Args:
    tensors: Tensors to ignore.
    op_info: Aggregated information about the op (e.g., quantization config).
    is_inbounding_tensor: Whether the tensors are the inbounding tensors for the
      op.

  Returns:
    A list of tensor transformation params for the ignored tensors.
  """
  op_ignored_tensor_params = []
  for tensor in tensors:
    tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
    no_quant_tensor_params = qtyping.OpToTensorParams(
        subgraph_op_id=op_info.subgraph_op_index,
        transformations=[qtyping.QuantTransformation.NO_QUANTIZE],
    )
    if is_inbounding_tensor:
      tensor_params = qtyping.TensorTransformationParams(
          tensor_name=tensor_name,
          consumers=[no_quant_tensor_params],
      )
    else:
      tensor_params = qtyping.TensorTransformationParams(
          tensor_name=tensor_name,
          producer=no_quant_tensor_params,
      )
    op_ignored_tensor_params.append(tensor_params)

  return op_ignored_tensor_params


def _merge_materialized_tensors(
    tensor_params: list[qtyping.TensorTransformationParams],
    ignored_input_tensor_params: Sequence[qtyping.TensorTransformationParams],
    ignored_output_tensor_params: Sequence[qtyping.TensorTransformationParams],
    op_info: qtyping.OpInfo,
    inputs_to_ignore: Sequence[int],
    outputs_to_ignore: Sequence[int],
) -> list[qtyping.TensorTransformationParams]:
  """Merge materialized tensors.

  Merge tensor transformation parameters for non-ignored and ignored tensors.
  The result list will keep the original order of inputs and outputs tensors in
  the op.

  Args:
    tensor_params: Tensor transformation params for non-ignored tensors in the
      op.
    ignored_input_tensor_params: Tensor transformation params for the ignored
      input tensors.
    ignored_output_tensor_params: Tensor transformation params for the ignored
      output tensors.
    op_info: Aggregated information about the op (e.g., quantization config).
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.

  Returns:
    Full list of transformation params for the op.
  """
  if not inputs_to_ignore and not outputs_to_ignore:
    return tensor_params

  result_tensor_params = []
  num_inputs = len([x for x in op_info.op.inputs if x != -1])
  num_outputs = len([x for x in op_info.op.outputs if x != -1])

  # Add input tensors.
  if inputs_to_ignore:
    input_idx, ignored_input_idx = 0, 0
    for i in range(num_inputs):
      if i in inputs_to_ignore:
        result_tensor_params.append(
            ignored_input_tensor_params[ignored_input_idx]
        )
        ignored_input_idx += 1
      else:
        result_tensor_params.append(tensor_params[input_idx])
        input_idx += 1
  else:
    result_tensor_params.extend(tensor_params[:num_inputs])

  # Add output tensors.
  output_start_idx = num_inputs - len(inputs_to_ignore)
  if outputs_to_ignore:
    output_idx, ignored_output_idx = output_start_idx, 0
    for i in range(num_outputs):
      if i in outputs_to_ignore:
        result_tensor_params.append(
            ignored_output_tensor_params[ignored_output_idx]
        )
        ignored_output_idx += 1
      else:
        result_tensor_params.append(tensor_params[output_idx])
        output_idx += 1
  else:
    result_tensor_params.extend(tensor_params[output_start_idx:])

  return result_tensor_params


def _tensor_indices_with_dtype(
    tensors: Sequence[int],
    subgraph_tensors: Sequence[schema_py_generated.TensorT],
    tensor_dtype_codes: Sequence[int],
) -> list[int]:
  """Get the indices of tensors with any of the given dtype.

  Args:
    tensors: A list of tensors (indices) from the subgraph.
    subgraph_tensors: A list of tensors in the subgraph.
    tensor_dtype_codes: A list of tensor dtype codes.

  Returns:
    A list of indices of tensors with the given dtype.
  """
  selected_indices = []
  for i, tensor_index in enumerate(tensors):
    tensor = subgraph_tensors[tensor_index]
    if tensor.type in tensor_dtype_codes:
      selected_indices.append(i)
  return selected_indices


def _add_non_match_tensors_to_ignored_lists(
    op: schema_py_generated.OperatorT,
    subgraph_tensors: Sequence[schema_py_generated.TensorT],
    dtypes_to_keep: Sequence[int],
    inputs_to_ignore: Sequence[int],
    outputs_to_ignore: Sequence[int],
) -> tuple[list[int], list[int]]:
  """Include tensors not matching the specified dtype in the ignored lists.

  Args:
    op: The op to be processed.
    subgraph_tensors: A list of tensors in the subgraph.
    dtypes_to_keep: A list of tensor dtype codes that need to be kept (not in
      the ignored lists).
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.

  Returns:
    A tuple of updated inputs_to_ignore and outputs_to_ignore.
  """
  input_indices = set(range(len(op.inputs)))
  inputs_to_keep = set(
      _tensor_indices_with_dtype(op.inputs, subgraph_tensors, dtypes_to_keep)
  )
  inputs_to_keep -= set(inputs_to_ignore)  # remove already ignored tensors.
  inputs_to_ignore = list(input_indices - inputs_to_keep)

  output_indices = set(range(len(op.outputs)))
  outputs_to_keep = set(
      _tensor_indices_with_dtype(op.outputs, subgraph_tensors, dtypes_to_keep)
  )
  outputs_to_keep -= set(outputs_to_ignore)  # remove already ignored tensors.
  outputs_to_ignore = list(output_indices - outputs_to_keep)
  return inputs_to_ignore, outputs_to_ignore


def _get_min_max_from_quant_params(
    quant_params: qtyping.UniformQuantParams,
) -> tuple[np.ndarray, np.ndarray]:
  """Recalculate min/max from tensor quantization params."""
  q_min, q_max = uniform_quantize_tensor.get_quantized_range(
      _IntType(quant_params.num_bits, True)
  )
  float_min = uniform_quantize_tensor.uniform_dequantize(
      np.array(q_min), quant_params
  )
  float_max = uniform_quantize_tensor.uniform_dequantize(
      np.array(q_max), quant_params
  )
  # We use qmax values to compute scale for symmetric quantization (see
  # uniform_quantize_tensor.tensor_zp_scale_from_min_max).
  if quant_params.symmetric:
    float_min = -float_max
  return float_min, float_max


def materialize_standard_op(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
    constraint: OpQuantConstraint = OpQuantConstraint.NO_CONSTRAIN,
    inputs_to_ignore: Optional[Sequence[int]] = None,
    outputs_to_ignore: Optional[Sequence[int]] = None,
) -> list[qtyping.TensorTransformationParams]:
  """Default materialization function for an op.

  Use materialize_fc_conv as the entry point to materialize FULLY_CONNECTED,
  CONV_2D, DEPTHWISE_CONV_2D as these ops may contain fused bias.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    constraint: The constraint for materializing the op.
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias). The returned list has the structure:
    [input_tensor_0_params, ..., input_tensor_n_params,
    output_tensor_0_params, ..., output_tensor_m_params].
  """
  inputs_to_ignore = inputs_to_ignore or []
  outputs_to_ignore = outputs_to_ignore or []
  # Filter out non-fp32 tensors (e.g., int32 indices).
  fp32_type_code = 0  # See schema_py_generated.py for type code.
  inputs_to_ignore, outputs_to_ignore = _add_non_match_tensors_to_ignored_lists(
      op_info.op,
      graph_info.subgraph_tensors,
      [fp32_type_code],
      inputs_to_ignore,
      outputs_to_ignore,
  )

  # Process op inputs and outputs.
  ignored_input_tensors, input_tensors, inputs_to_ignore = (
      _split_tensors_by_indices(
          op_info, graph_info, inputs_to_ignore, is_inbounding_tensor=True
      )
  )
  ignored_output_tensors, output_tensors, outputs_to_ignore = (
      _split_tensors_by_indices(
          op_info, graph_info, outputs_to_ignore, is_inbounding_tensor=False
      )
  )
  # Materialize op tensors.
  if not input_tensors and not output_tensors:
    tensor_params = []  # Every tensor is ignored.
  elif constraint == OpQuantConstraint.SAME_AS_INPUT_SCALE:
    tensor_params = _materialize_standard_op_with_same_as_input_scale(
        input_tensors,
        output_tensors,
        op_info,
        graph_info,
        tensor_name_to_qsv,
        get_tensor_quant_params_fn,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )
  elif constraint == OpQuantConstraint.SAME_AS_OUTPUT_SCALE:
    tensor_params = _materialize_standard_op_with_same_as_output_scale(
        input_tensors,
        output_tensors,
        op_info,
        graph_info,
        tensor_name_to_qsv,
        get_tensor_quant_params_fn,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )
  else:
    tensor_params = _materialize_standard_op_no_constraint(
        input_tensors,
        output_tensors,
        op_info,
        graph_info,
        tensor_name_to_qsv,
        get_tensor_quant_params_fn,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )

  # Materialize ignored tensors.
  ignored_input_tensor_params = _materialize_ignored_tensors(
      ignored_input_tensors, op_info, is_inbounding_tensor=True
  )
  ignored_output_tensor_params = _materialize_ignored_tensors(
      ignored_output_tensors, op_info, is_inbounding_tensor=False
  )
  # Combine all tensor params keeping the original order.
  return _merge_materialized_tensors(
      tensor_params,
      ignored_input_tensor_params,
      ignored_output_tensor_params,
      op_info,
      inputs_to_ignore,
      outputs_to_ignore,
  )


def materialize_op_with_output_activation_constraint(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    output_activation_constraints: dict[int, qtyping.UniformQuantParams],
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    tensor_quant_params_cache: TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in an op with output activation constraint.

  The output activation constraints are used to explicitly set
  quantization parameters for the output tensor when doing SRQ.

  Function assumes that there is only one output tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    output_activation_constraints: A map of output activation num bits to
      quantization parameters.
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).

  Raises:
    ValueError: If the op has more than one output tensor, or if the output
      activation constraints dictionary does not contain an entry for the
      activation num bits specified in the op quantization config.
  """
  if len(op_info.op.outputs) != 1:
    raise ValueError(
        "Materialize op with output activation constraint only supports ops"
        " with a single output tensor."
    )

  tensor_params = materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )
  output_tensor_params = tensor_params[-1]

  # Explicitly set quantization parameters for the output tensor when doing SRQ.
  activation_tensor_config = op_info.op_quant_config.activation_tensor_config
  if (
      activation_tensor_config is not None
      and output_tensor_params.producer is not None
  ):
    activation_num_bits = activation_tensor_config.num_bits
    if activation_num_bits not in output_activation_constraints:
      raise ValueError(
          "Output activation constraints dictionary does not contain entity"
          f" for activation num bits {activation_num_bits}."
      )
    fixed_quant_params = output_activation_constraints[activation_num_bits]
    op_tensor_params = qtyping.OpToTensorParams(
        subgraph_op_id=output_tensor_params.producer.subgraph_op_id,
        transformations=output_tensor_params.producer.transformations,
        parameters=fixed_quant_params,
    )
    output_tensor_params.producer = op_tensor_params
    # Update the tensor_name_to_qsv map using the output activation constraints.
    min_val, max_val = _get_min_max_from_quant_params(
        fixed_quant_params,
    )
    tensor_name_to_qsv[output_tensor_params.tensor_name]["min"] = min_val
    tensor_name_to_qsv[output_tensor_params.tensor_name]["max"] = max_val

  return tensor_params


def get_tensor_transformations(
    op_quant_config: qtyping.OpQuantizationConfig,
    is_inbounding_tensor: bool,
    is_constant: bool,
):
  """Get the transformations for the tensor.

  Args:
    op_quant_config: the quantization config for the op.
    is_inbounding_tensor: whether the tensor is an inbounding tensor for the op.
    is_constant: whether the tensor is a constant tensor.

  Returns:
    The transformations for the tensor.
  """
  # Check if SRQ.
  if (
      op_quant_config.compute_precision == qtyping.ComputePrecision.INTEGER
      and op_quant_config.activation_tensor_config is not None
  ):
    if is_inbounding_tensor:
      transformations = [_QuantTransformation.ADD_QUANTIZE]
      if is_constant:
        # Quantize the constant tensor directly to simplify downstream
        # optimizations.
        transformations = [_QuantTransformation.QUANTIZE_TENSOR]
    else:
      transformations = [_QuantTransformation.ADD_DEQUANTIZE]
  # Check if DRQ.
  elif (
      op_quant_config.compute_precision == qtyping.ComputePrecision.INTEGER
      and op_quant_config.activation_tensor_config is None
  ):
    if is_inbounding_tensor and is_constant:
      transformations = [_QuantTransformation.QUANTIZE_TENSOR]
    else:
      transformations = [_QuantTransformation.NO_QUANTIZE]
  # Check if WEIGHT_ONLY.
  elif (
      op_quant_config.compute_precision == qtyping.ComputePrecision.FLOAT
      and op_quant_config.explicit_dequantize
  ):
    if is_inbounding_tensor and is_constant:
      # ADD_DEQUANTIZE is always accompanined with a quantization parameters.
      # Thus [ADD_DEQUANTIZE] is equivalent to [QUANTIZE_TENSOR, ADD_DEQUANTIZE]
      # downstream pattern: quantized_tensor -> dequantize op -> float_tensor.
      transformations = [_QuantTransformation.ADD_DEQUANTIZE]
    else:
      transformations = [_QuantTransformation.NO_QUANTIZE]
  else:
    raise ValueError(
        "Unsupported compute precision: %s" % op_quant_config.compute_precision
    )
  return transformations


def get_tensor_transformation_params(
    tensor_name: str,
    op_info: qtyping.OpInfo,
    is_inbounding_tensor: bool,
    quant_params: Optional[qtyping.UniformQuantParams] = None,
    is_constant: bool = False,
) -> qtyping.TensorTransformationParams:
  """Transformation params for the op's tensor.

  Args:
    tensor_name: the name of the tensor.
    op_info: aggregated information about the op (e.g., quantization config).
    is_inbounding_tensor: whether the tensor is inbounding tensor to the op.
    quant_params: the quantization parameters for the tensor.
    is_constant: whether the tensor is a constant tensor.

  Returns:
    The transformation for the op's tensor.
  """
  transformations = get_tensor_transformations(
      op_info.op_quant_config, is_inbounding_tensor, is_constant
  )
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=transformations,
  )
  if is_inbounding_tensor:
    return qtyping.TensorTransformationParams(
        tensor_name=tensor_name,
        consumers=[op2tensor_params],
    )
  return qtyping.TensorTransformationParams(
      tensor_name=tensor_name,
      producer=op2tensor_params,
  )


def get_weight_quantized_dim(
    op_info: qtyping.OpInfo,
    tensor_data: np.ndarray,
    granularity: qtyping.QuantGranularity,
):
  """Get the quantized dimension for the weight tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_data: The weight tensor data.
    granularity: The granularity of the weight tensor.

  Returns:
    The quantized dimension for the weight tensor.
  """
  quantized_dim = None
  if granularity == qtyping.QuantGranularity.CHANNELWISE:
    if op_info.op_name == _TFLOpName.BATCH_MATMUL:
      quantized_dim = get_bmm_weight_quantized_dim(
          tensor_data, adj_y=op_info.op.builtinOptions.adjY
      )
    else:
      quantized_dim = tfl_flatbuffer_utils.TFL_OP_TO_WEIGHT_QUANTIZED_DIM.get(
          op_info.op_name, None
      )
  elif uniform_quantize_tensor.is_blockwise(granularity):
    quantized_dim = (
        tfl_flatbuffer_utils.TFL_OP_TO_BLOCKWISE_WEIGHT_QUANTIZED_DIM[
            op_info.op_name
        ]
    )
  return quantized_dim


def get_reduce_dims(
    quantized_dim: Optional[int],
    tensor_shape: Sequence[int],
) -> Optional[tuple[int, ...]]:
  """Get the reduce dims of a tensor for the given quantized dimension."""
  if quantized_dim is None:
    return None
  reduce_dims = []
  for rank_idx in range(len(tensor_shape)):
    if rank_idx != quantized_dim:
      reduce_dims.append(rank_idx)
  return tuple(reduce_dims)


def get_bmm_weight_quantized_dim(
    weight_tensor_data: np.ndarray, adj_y: bool
) -> int:
  """Get the quantized dimension for batch matmul."""
  rank = len(weight_tensor_data.shape)
  # If adj_y is true, the weight tensor is transposed.
  if adj_y:
    return rank - 2
  return rank - 1



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with fully connected op."""

import pathlib

from absl.testing import parameterized
import absl.testing.absltest as absltest

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils


_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig
_AlgorithmName = quantizer.AlgorithmName

_TEST_MODEL_FOLDER = test_utils.get_path_to_datafile(
    '../models/dequantized_weights/'
)


# TODO: b/398335637 - Better ways to construct op configs for different cases.
class FullyConnectedTest(test_utils.BaseOpTestCase):

  def setUp(self):
    super().setUp()
    self._op_name = _OpName.FULLY_CONNECTED

  @parameterized.product(
      # (algorithm_name, weight_tolerance, output_tolerance)
      algorithm_and_tolerances=[
          (_AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY, 1e-5, 1e-5),
          (_AlgorithmName.MIN_MAX_UNIFORM_QUANT, 1e-2, 1e-1),
          (_AlgorithmName.OCTAV, 1e-3, 1e-1),
          # Only check size reduction for MSE.
          (_AlgorithmName.MSE, 10, 10),
      ],
      # (model_name, granularity)
      model_and_granularity=[
          (
              'tensor_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.TENSORWISE,
          ),
          (
              'channel_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.CHANNELWISE,
          ),
      ],
      weight_bit_width=[4, 8],
  )
  def test_weight_only_quantization_accuracy_and_size_within_tolerance(
      self, algorithm_and_tolerances, model_and_granularity, weight_bit_width
  ):
    algorithm_key, weight_tolerance, output_tolerance = algorithm_and_tolerances
    model, granularity = model_and_granularity
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_bit_width,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.FLOAT,
        explicit_dequantize=True,
    )
    expected_model_size_reduction = (
        80 if op_config.weight_tensor_config.num_bits == 4 else 65
    )
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )

  @parameterized.product(
      algorithms=[
          _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          _AlgorithmName.OCTAV,
      ],
      granularity=[
          qtyping.QuantGranularity.BLOCKWISE_32,
          qtyping.QuantGranularity.BLOCKWISE_64,
          qtyping.QuantGranularity.BLOCKWISE_128,
      ],
  )
  def test_int4_blockwise_weight_only_quantization_raise_error(
      self, algorithms, granularity
  ):
    algorithm_key = algorithms
    model = 'tensor_i4rangedvalues_fc.tflite'
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=4,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.FLOAT,
        explicit_dequantize=True,
    )

    # Blockwise quantization is not supported for weight-only quantization.
    error_msg = 'Unsupported op'
    with self.assertRaisesRegex(ValueError, error_msg):
      self.quantize_and_validate(
          model_path, algorithm_key, self._op_name, op_config
      )

  @parameterized.product(
      # (algorithm_name, weight_tolerance, output_tolerance)
      algorithm_and_tolerances=[
          (_AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY, 1e-5, 1e-2),
          (_AlgorithmName.MIN_MAX_UNIFORM_QUANT, 1e-2, 1e-1),
          (_AlgorithmName.OCTAV, 1e-3, 1e-1),
          (_AlgorithmName.MSE, 10, 10),
      ],
      # (model_name, granularity)
      model_and_granularity=[
          (
              'tensor_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.TENSORWISE,
          ),
          (
              'channel_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.CHANNELWISE,
          ),
      ],
      weight_bit_width=[4, 8],
  )
  def test_dynamic_quantization_accuracy_and_size_within_tolerance(
      self, algorithm_and_tolerances, model_and_granularity, weight_bit_width
  ):

    algorithm_key, weight_tolerance, output_tolerance = algorithm_and_tolerances
    model, granularity = model_and_granularity
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_bit_width,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )
    expected_model_size_reduction = (
        80 if op_config.weight_tensor_config.num_bits == 4 else 65
    )
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )

  @parameterized.product(
      algorithms=[
          _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          _AlgorithmName.OCTAV,
      ],
      granularity=[
          qtyping.QuantGranularity.BLOCKWISE_32,
          qtyping.QuantGranularity.BLOCKWISE_64,
          qtyping.QuantGranularity.BLOCKWISE_128,
      ],
  )
  def test_int4_blockwise_dynamic_quantization_accuracy_and_size_within_tolerance(
      self, algorithms, granularity
  ):
    algorithm_key = algorithms
    weight_tolerance = 1e-2
    output_tolerance = 1e-1
    model = 'tensor_i4rangedvalues_fc.tflite'
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=4,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )

    expected_model_size_reduction = 65
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )

  @parameterized.product(
      # (algorithm_name, weight_tolerance, output_tolerance)
      algorithm_and_tolerances=[
          (_AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY, 1e-5, 1e-1),
          (_AlgorithmName.MIN_MAX_UNIFORM_QUANT, 1e-2, 1e-1),
          (_AlgorithmName.OCTAV, 1e-3, 1e-1),
          # Only check size reduction for MSE.
          (_AlgorithmName.MSE, 10, 10),
      ],
      # (model_name, granularity)
      model_and_granularity=[
          (
              'tensor_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.TENSORWISE,
          ),
          (
              'channel_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.CHANNELWISE,
          ),
      ],
      weight_bit_width=[4, 8],
  )
  def test_a8static_quantization_accuracy_and_size_within_tolerance(
      self, algorithm_and_tolerances, model_and_granularity, weight_bit_width
  ):
    algorithm_key, weight_tolerance, output_tolerance = algorithm_and_tolerances
    model, granularity = model_and_granularity
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=8,
            symmetric=False,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_bit_width,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )
    expected_model_size_reduction = (
        80 if op_config.weight_tensor_config.num_bits == 4 else 65
    )
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )

  @parameterized.product(
      # (algorithm_name, weight_tolerance, output_tolerance)
      algorithm_and_tolerances=[
          (_AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY, 1e-5, 1e-1),
          (_AlgorithmName.MIN_MAX_UNIFORM_QUANT, 1e-2, 1e-1),
          (_AlgorithmName.OCTAV, 1e-3, 1e-1),
          # Only check size reduction for MSE.
          (_AlgorithmName.MSE, 10, 10),
      ],
      # (model_name, granularity)
      model_and_granularity=[
          (
              'tensor_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.TENSORWISE,
          ),
          (
              'channel_i4rangedvalues_fc.tflite',
              qtyping.QuantGranularity.CHANNELWISE,
          ),
      ],
      weight_bit_width=[4, 8],
  )
  def test_a16static_quantization_accuracy_and_size_within_tolerance(
      self, algorithm_and_tolerances, model_and_granularity, weight_bit_width
  ):
    algorithm_key, weight_tolerance, output_tolerance = algorithm_and_tolerances
    model, granularity = model_and_granularity
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=16,
            symmetric=True,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_bit_width,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )
    expected_model_size_reduction = (
        80 if op_config.weight_tensor_config.num_bits == 4 else 65
    )
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )

  @parameterized.product(
      algorithms=[
          _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          _AlgorithmName.OCTAV,
      ],
      granularity=[
          qtyping.QuantGranularity.BLOCKWISE_32,
          qtyping.QuantGranularity.BLOCKWISE_64,
          qtyping.QuantGranularity.BLOCKWISE_128,
      ],
      activation_bit_width=[8, 16],
  )
  def test_int4_blockwise_static_quantization_raise_error(
      self, algorithms, granularity, activation_bit_width
  ):
    algorithm_key = algorithms
    model = 'tensor_i4rangedvalues_fc.tflite'
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model
    op_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=activation_bit_width,
            symmetric=True,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=4,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )

    # Blockwise quantization is not supported for static quantization.
    error_msg = 'Unsupported op'
    with self.assertRaisesRegex(ValueError, error_msg):
      self.quantize_and_validate(
          model_path, algorithm_key, self._op_name, op_config
      )

  @parameterized.product(
      weight_bit_width=[4, 8],
      algorithm=[
          _AlgorithmName.HADAMARD_ROTATION,
          _AlgorithmName.DECOMPOSED_HADAMARD_ROTATION,
      ],
  )
  def test_hadamard_rotation_accuracy_and_size_within_tolerance(
      self, weight_bit_width, algorithm
  ):
    algorithm_key = algorithm
    # TODO(b/417309114): Soft skip weight errors because they're rotated hence
    # not expected to match.
    weight_tolerance = 1
    output_tolerance = 2e-2
    model_path = test_utils.get_path_to_datafile(
        '../models/conv_fc_mnist.tflite'
    )
    granularity = qtyping.QuantGranularity.CHANNELWISE
    op_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_bit_width,
            symmetric=True,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,
        explicit_dequantize=False,
    )
    expected_model_size_reduction = (
        80 if op_config.weight_tensor_config.num_bits == 4 else 65
    )
    self.assert_quantization_accuracy_and_size(
        algorithm_key,
        model_path,
        self._op_name,
        op_config,
        expected_model_size_reduction,
        weight_tolerance,
        output_tolerance,
    )


if __name__ == '__main__':
  absltest.main()
